{ Unit Debug.pas

  This unit is the center of the debugging process.
  It allows to drive the emulator EMU48.

  In this unit, you will find the singleton class TDebugger
  that is the interface between the software and the debugger.
  The Debugger class provide directly the folowing operations:

  operations linked to the emulator
    Run an emulator
    Stop the emulator
    Step in in ASM                
    Step in in RPL
    Step Over in ASM
    Step Over in RPL
    Continue ASM
    Continue RPL
    Breakpoint management
    Call Back on RPL or ASM instruction
    Read the memory
    access to the emulated chipset.

  very usefull operations
    Edit the source at an address
    Edit the source at a symbol


  The TDebugger class uses a child of TLDeb to
  comunicate with the emulator itself. TLDeb providing a standard
  interface over the emulator.

  Two childs of the TLDeb class are provided: one for EMU48, one for YORKEM    // wgg dropped Yorkem

  The breakpoints are childs of the TMark object of the editor in
  order to be able to see the breakpoints in the gutter of the
  editor. The Editor has to be indicated to the TDebugger through the
  Editor property.

  This unit is copyright Cyrille de Brebisson and Hewlett Packard

  Changes marked with WGG are copyright 2002, William G. Graves

}
unit Debug;

interface
uses
  windows, classes, sysutils, lrworking;

const
  DOCOL = $02D9D;         // This is used by the RPL Breakpoint engine
                          // in order to put a rpl breakpoint that falls
                          // on a program 5 nibble later
                          // we need to do this because the DOCOL
                          // is not executed directly.
  Seconde1 = 1/24/60/60;
  DebuggerMessage = 7000;
  NbLastInstr = 255;      // This constant allows to specify the number of
                          // instructions to save. For some obvious reasons,
                          // this must be a power of 2-1

type

  // This event is used by the debugger to prompt for the source file
  // corresponding to a specified address
  TEditAdr = procedure (sender: TObject; adr: integer; clear: boolean=False) of object;        // wgg
  // call backs on RPL instruction
  TRplInstr = Procedure (Sender: TObject; D0, D1, B: integer) of object;
  // call backs on ASM instruction
  TAsmInstr = Procedure (Sender: TObject) of object;

  // An emulator type
  TDebType = (DTNone, DtEmu49);

  // Callbacks functions used by the TLDeb to comunicate with the TDebugger.
  TNotifyFunction = function (Sender: TObject): Boolean of object;
  TNotifyRplFunction = function (Sender: TObject; D0: integer): Boolean of object;

  // This is the YORKE structure
  TIChipset = record
    A, B, C, D, R0, R1, R2, R3, R4 : int64;
    D0, D1, PC : integer;
    RSTK : array [0..7] of integer;
    st : array [0..15] of boolean;
    hst : array [0..3] of boolean;
    carry, Inte, Decimal: boolean;
    P: integer;
    cycles: int64;
    CRC: integer;
    RomV1, RomV2: integer;
  end;
             
  // structure de sauvegarde des dernieres instructions
  TLinstr = array [0..NbLastInstr] of integer;

  // this is an abstract class that is used to
  // derivate an emulator interface object
  // This object handle the ASMBreakpoints but not the RPLstuff
  TLDeb = class (TObject)
  Private
    FAsmInstr : TNotifyFunction;
    FRplInstr : TNotifyRplFunction;
    FBye: TNotifyEvent;
    CalcModel: Char;                            // wgg X=HP49, G=HP48
  public
    // the 3 parameters are the callbacks when
    // an ASM Instruction is evaluated
    // a RPL instruction is evaluated
    // the emulator is closed
    Constructor Create(AsmInstr: TNotifyFunction; RplInstr: TNotifyRplFunction;
                bye: TNotifyEvent; Ram: TFileName; Port2: TFileName); virtual;
    // memory access function
    Procedure ReadMemory(var PByte: Array of byte; adress, size: integer); virtual; abstract;
    Procedure WriteMemory(var PByte: Array of byte; adress, size: integer); virtual; abstract;
    Function RomRead(a: integer): integer;        virtual; abstract;
    // switch the flash rom
    Procedure Switch(adr: integer);               virtual; abstract;
    // control the engine
    Procedure Run;                                virtual; abstract;
    Procedure Stop;                               virtual; abstract;
    Procedure StepIn;                             virtual; abstract;
    Procedure StepOver(adr: integer);             virtual; abstract;
    Function AsmDebug:boolean;                    virtual; abstract;
    procedure DebugInterrupt(value: boolean);     virtual;
    // chipset interface
    Procedure GetChipset(var Chipset: TIChipset); virtual; abstract;
    Procedure SetChipset(Chipset: TIChipset);     virtual; abstract;
    // asm breakpoint list management
    Procedure SetBreakList(b: array of integer);  virtual; abstract;
    Function DType: TDebType;                     virtual; abstract;
    Function LastAsmInstr: TLinstr;               virtual; abstract;
    Procedure LoadFile(f: TFileName);             virtual; abstract;
    Procedure ExecKey(s: string; action: TThreadMethod);                 virtual; abstract;
    function WaitShutDn: boolean;                 virtual; abstract;      // wgg
    function WaitBusy: boolean;                   virtual; abstract;      // wgg
    Function EmDestroy:boolean;                   virtual; abstract;
  Property CalcType: Char read CalcModel;
  end;

  // first child: the Emu48 emulator as the HP49
  TEmu49Deb = class (TLdeb)
  private
    StepByStep: boolean;
    AsmStepOverBreakPoint: integer;
    blist: array of integer;
    lasti: TLinstr;
    Procedure AsmInstruction(sender: tobject);
    Procedure RplInstruction(sender: tobject);
  public
    AutoClose: boolean;
    Constructor Create(AsmInstr: TNotifyFunction; RplInstr: TNotifyRplFunction;
                bye: TNotifyEvent; Ram: TFileName; Port2: TFileName); override;  // wgg
    destructor Destroy; override;
    Procedure ReadMemory(var PByte: Array of byte; adress, size: integer); Override;
    Procedure WriteMemory(var PByte: Array of byte; adress, size: integer); Override;
    Procedure Run;                                Override;
    Procedure Stop;                               Override;
    Procedure StepIn;                             Override;
    Procedure StepOver(adr: integer);             Override;
    procedure DebugInterrupt(value: boolean);     override;
    Function AsmDebug:boolean;                    Override;
    Procedure Switch(adr: integer);               Override;
    Procedure GetChipset(var Chipset: TIChipset); Override;
    Procedure SetChipset(Chipset: TIChipset);     Override;
    Procedure SetBreakList(b: array of integer);  Override;
    Function RomRead(a: integer): integer;        Override;
    Function DType: TDebType;                     Override;
    function LastAsmInstr: TLinstr;               Override;
    Procedure LoadFile(f: TFileName);             Override;
    Procedure ExecKey(s: string; action: TThreadMethod);          Override;
    function WaitShutDn: boolean;                 Override;                  // wgg
    function WaitBusy: boolean;                   Override;                  // wgg
    Function EmDestroy:boolean;                   Override;
  end;

  // This object is a breakpoint.
  // is derivated from the TMark of the Editor in order to
  // be visible as a mark in the gutter.
  // it records every important data of a breakpoint.
  TBreakPoint = class (TObject)
  public
    Active: Boolean;
    Adresse: integer;
    rpl: boolean;
    fil: TLFile;
    FileName: TFileName;
    line: integer;
    constructor Create;
  end;

// wgg a breakpoint in a source code is file, line number
// used until we can reconstruct the object
  TSrceBreakPoint = class (TObject)
  public
    fil: TLFile;
    FileName: TFileName;
    line: integer;

  {  constructor Create; }
  end;

  flashptr = record
    value: integer;
    calls: integer;
    name: string;
  end;

  TProfilerItem = record
    D0: integer;
    time: integer;
    nbcalls: integer;
  end;

  TProfilerStack = record
//    StackSize: integer;
    timein: integer;
    D0: integer;
  end;

  // This object is the heart of the whole thing.
  // They must be only one instance of this object in the system at a time!
  {$HINTS OFF}
  TDebugger = class (TComponent)
  private
    FlashPtrList: array [0..2000] of FlashPtr; FlashPtrUsed: integer;
    FActiveFlashPtr: boolean;          // true if the flash ptr are recorded
    RplStepOverBreakPoint: Integer;   // This is the address for the end of the RPL step Over process
    BL: TList;                        // the breakpoint list
    BSL: array of TPos;               // a temporary list of TPOS for source breakpoints
    FCallBackEdit: TEditAdr;
    FOnNewBreak: TNotifyEvent;
    FOnStateChange: TNotifyEvent;
    FOnAsmInstr, FOnRplInstr: THandle; // when an RPL or ASM instruction fires, the debugger
                                       // send a message DebuggerMessage to the specify handle.
                                       // Thus allowing the instruction to be processed in the main thread
    NbRplBreakPoints: integer;         // This contains a list of the RPL breakpoints. This is recreated
    RplBreakPoints: array [0..1023] of integer; // each time the break point list changes
    FChipset: TIChipset;               // contains a copy of the last known chipset
    deb: TLDeb;                        // the debuging object
    RplFullSpeed: boolean;             // do not break on RPL instructions
    FEraseStk: boolean;
    now: int64;                        // This time betwee 2 RPL instruction is recorded, therfore, I must keep the last time
    LastRPLInstr: TLinstr;             // a table of the last RPL instrctions
    LastRplPointer: integer;           // pointer on the last rpl circular buffer.
    FRamFile: string;
    FPort2File: string;                // wgg
    FProfiler: boolean;
    ProfilerList: array of TProfilerItem;
    ProfilerUsed: integer;
    ProfilerStack: array of TProfilerStack;
    ProfilerStackPointer: integer;
    ProfilerStart: int64;
    FDebugInt: Boolean;
    procedure SetOnStateChange(const Value: TNotifyEvent);
    Function AsmInstruction(Sender: TObject): Boolean;              // Used to dialog with the debugger
    Function RplInstruction(Sender: TObject; d0: integer): boolean; // Used to dialog with the debugger
    Procedure DebBye(Sender: TObject);                              // Used to dialog with the debugger
    function GetBList(i: integer): TBreakPoint;                     // Get an element of the breakpoint list
    function GetRunning: Boolean;
    procedure SetDebugType(const Value: TDebType);
    function GetDebugType: TDebType;
    procedure continue;                                             // let the emulation continue freeley
    function GLastInstr: TLinstr;
    procedure SetEraseStk(const Value: boolean);
    function GLastRplInstr: TLinstr;
    constructor Create(aowner: TComponent); override;               // this method is here on purpose: In order to avoid creation of more than 1 debugger.
    Procedure SetActiveFlashPtr(Value: boolean);
    Procedure SetProfiler(b: boolean);
    procedure SetDebugInt(const Value: Boolean);
  public
    RplTime: int64;
    ZeroReturn: boolean;
    TestMemMode: Boolean;
    TestMemAdr: integer;
    FLASHPSEC: integer;                                    // wgg
    TestMemValue, TestMemMask: array of byte;
    NextRplStepOver: integer;
    Property EraseStk: boolean read FEraseStk write SetEraseStk;
    Property ActiveFlashPtr: Boolean read FActiveFlashPtr write SetActiveFlashPtr;
    Function FlashPtrResult: String;
    property Profiler: boolean read FProfiler write SetProfiler;
    Function ProfilerResult: string;
    function GetChipset: TIChipset;               // wgg made public for frmDebugEMU   // update the FChipset structure
    function InAsmStep: boolean;                  // wgg added for ASM Debug window, and block a build
    function InRplStep: boolean;                  // wgg added for Rpl block a build
    Destructor Destroy; override;

    Procedure StopAsm;
    Procedure StopRpl;
    Procedure continueAsm;
    Procedure continueRpl;
    Procedure StepOverAsm;
    Procedure StepOverRpl;
    Procedure StepInAsm;
    Procedure StepInRpl;
    Procedure StepIn;
    Procedure StepOver;

    Procedure ReadMemory(var PByte: Array of byte; adress, size: integer);
    Procedure WriteMemory(var PByte: Array of byte; adress, size: integer);
    Function ReadInteger(Adr: integer): integer;
    Procedure WriteInteger(adr: integer; value: integer);
    Function ReadInt64(Adr: Integer): int64;
    Function FullAdr(adr: integer): integer;
    Function ReadRom(adr: integer): integer;
    Procedure RemoveAllBreakPoints;                                           // wgg made public
    Function SetBreakPoint(Adress: integer; Rpl: Boolean; active: boolean = true): integer;
    procedure recomputeBP;                                                    //  wgg
    procedure SaveBPSource;                                                   //  wgg
    Procedure RemoveBreakPoint(i: integer);
    function GetBreakPointIndex(Adress: integer): integer;                    // wgg
    Procedure ListeBreakpoints;
    function BreakCount: integer;
    Property BList[i: integer]: TBreakPoint read GetBList;
    property OnNewBreak: TNotifyEvent read FOnNewBreak write FOnNewBreak;
    Procedure RebuiltBreakPointList;
    Procedure EditAtAddress(a: integer; doBar:Boolean=False);                  // wgg
    Procedure EditAtSymbol(s: string; doBar:Boolean=False);                    // wgg
    property Running: Boolean read GetRunning;
    property OnAsmInstr: THandle read FOnAsmInstr write FOnAsmInstr;
    property OnRplInstr: THandle read FOnRplInstr write FOnRplInstr;
    property DebugInt: Boolean read FDebugInt write SetDebugInt;
    property OnEditAdr: TEditAdr read FCallBackEdit write FCallBackEdit;
    Property OnStateChange: TNotifyEvent read FOnStateChange write SetOnStateChange;
    property Chipset: TIChipset read FChipset;
    Procedure SaveChipset;
    property DebugType: TDebType read GetDebugType write SetDebugType;
    property lastInstrs : TLinstr read GLastInstr;
    property lastRplInstrs: TLinstr read GLastRplInstr;
    Procedure LoadFile(f: TFileName);
    Property RamFile: string read FRamFile write FRamFile;
    Property Port2File: string read FPort2File write FPort2File;               // wgg
    function CalcModel: char;                                                  // wgg
    Procedure ExecKey(s: string; action: TThreadMethod);
    function WaitShutDn: boolean;                                              // wgg
    function WaitBusy: boolean;                                                // wgg
    function crntAdrs: integer;
  end;
  {$HINTS ON}

var
  Debugger: TDebugger;

implementation

uses forms, fmBreakPointChoice, idGlobal, FmProject, fmDebugRPL, fmSourceEdit,
  fmDebugEMU, emu48dll;     // wgg

{ TDebugger }

function TDebugger.CalcModel: char;                                           // wgg
begin
  if Deb <> Nil then result:= Deb.CalcModel
  else result:='^';                              // wgg we use as an error signal
end;

function TDebugger.BreakCount: integer;
begin
  result:= bl.count;
end;

procedure TDebugger.continue;
begin
  if Deb<>nil then
    Deb.Run;
end;

Procedure TDebugger.SetActiveFlashPtr(Value: boolean);
Begin
  if Value=FActiveFlashPtr then exit;
  FActiveFlashPtr:= Value;
  if FActiveFlashPtr then
    FlashPtrUsed:= 0;
end;

Function TDebugger.FlashPtrResult: String;
var
  i, j: integer;
  a: flashptr;
  s: string;
Begin
  // bubble sort!
  for i:=0 to FlashPtrUsed-1 do
    for j:=0 to FlashPtrUsed-2 do
      if FlashPtrList[j].calls<FlashPtrList[j+1].calls then
        begin a:= FlashPtrList[j]; FlashPtrList[j]:= FlashPtrList[j+1]; FlashPtrList[j+1]:= a; end;

  result:= '';
  for i:=0 to FlashPtrUsed-1 do
  Begin
    s:= SymbolTable.GetFromAdr2(FlashPtrList[i].value);
    if (s='') or (s[1]<>'^') then
    Begin
      s:= SymbolTable.GetFromAdr2(SymbolTable.ValueFromText('FlashList'+inttostr((FlashPtrList[i].value+1) and $F))+
                                  (FlashPtrList[i].value shr 12)-1);
      if s='' then
        s:= IntToHex(FlashPtrList[i].value,7);
    end;
    s:= copy(s+'                          ', 1, 20);
    result:= result+s+inttostr(FlashPtrList[i].calls)+#13#10;
  end;
end;

constructor TDebugger.Create(aowner: TComponent);
begin
  inherited create(aowner);
  bl:= TList.Create;    // create the breakpoint list
  DebugInt:= False;
  EraseStk:= False;
end;

Procedure TDebugger.SetProfiler(b: boolean);
var
  i: integer;
Begin
  if b then
  begin
    ProfilerUsed:= 0;
    ProfilerStackPointer:= 0;
    FProfiler:= true;
    GetChipset;
    ProfilerStart:= Chipset.cycles;
    for i:=0 to high(ProfilerStack) do with ProfilerStack[i] do begin D0:= 0; TimeIn:= 0; end;
    // get i = RSTK size
//    GetChipset;
//    i:= 0; j:= Chipset.B-5; while ReadInteger(j)<>0 do begin inc(i); dec(j); end;
//    if length(ProfilerStack)<i then SetLength(ProfilerStack, i+20);
//    ProfilerStackPointer:= i-1;
//    GetChipset;
//    while i<>0 do with ProfilerStack[i-1] do begin D0:=-1; timein:= Chipset.cycles; dec(i); end;
  end else
  begin
    FProfiler:= false;
    GetChipset;
    ProfilerStart:= Chipset.cycles-ProfilerStart;
  end;
end;

Function TDebugger.ProfilerResult: string;
var
  i, j, k: integer;
  a: TProfilerItem;
  s: string;
Begin
  // bubble sort!
  for i:=0 to ProfilerUsed-1 do
    for j:=0 to ProfilerUsed-2 do
      if ProfilerList[j].time<ProfilerList[j+1].time then
        begin a:= ProfilerList[j]; ProfilerList[j]:= ProfilerList[j+1]; ProfilerList[j+1]:= a; end;

  k:=0;
  result:= '';
  for i:=0 to ProfilerUsed-1 do
  Begin
    s:= SymbolTable.GetFromAdr2(ProfilerList[i].d0-10);
    if (s='') then
      s:= inttohex(ProfilerList[i].d0,6);
    s:= copy(s+'                          ', 1, 20);
    result:= result+s+
                 copy(floattostr(ProfilerList[i].time / 4000)+'ms              ', 1, 15)+
                 inttostr(ProfilerList[i].nbcalls)+#13#10;
    inc(k,ProfilerList[i].time);
  end;
  result:=result+#13#10#13#10'Sum: '+floattostr(k/4000)+'ms'#13#10'Elapsed: '+floattostr(ProfilerStart/4000)+'ms';

end;

// This function is called by the emulator each time a RPL instruction is
// executed
Function TDebugger.RplInstruction(Sender: TObject; D0: integer): boolean;
var
  i, j, D02: integer;
  b: boolean;
  a: array of byte;
begin

  { This is the memory change test function
    This peace of code allows you to control
    the memory after each RPL instruction in order
    to check a perticular area of memory.
    I do not realy remember how it work, so help yourself! }
  if not ((TestMemAdr=0) or (TestMemAdr=-1)) then
  Begin
    setlength(a,Length(TestMemValue));
    Deb.ReadMemory(a,TestMemAdr,Length(TestMemValue));
    b:= true;
    for i:= 0 to Length(TestMemValue)-1 do
      b:= b and ( (a[i] and TestMemMask[i]) = TestMemValue[i]);
    if (TestMemMode and b) or ((not TestMemMode) and (not b)) then
    Begin
      RplFullSpeed:= False;
      NbRplBreakPoints:= 0;
      Application.MessageBox('Memory changed. Break into RPL Debugger.', 'Attention!', MB_OK);
    end;
  end;

  GetChipset;          // get the chipset

  { profiler stuff }
  if Profiler then
  try
  Begin
    // get i = RSTK size
    i:= ((Chipset.b and $FFFFF) - ReadInteger(SymbolTable.ValueFromText('RSKTOP'))) div 5;   // wgg
    // if the RSTK size changed
    if (i>0) and (i<100) and (chipset.D1>(chipset.b and $FFFFF)) and
      (i<>ProfilerStackPointer) then
      // if the size increased, we entered a sub function, else we are leaving one.
      if i>ProfilerStackPointer then
      Begin
        if length(ProfilerStack)<i then SetLength(ProfilerStack, i+20);  // stack size ok?
        for j:=ProfilerStackPointer+1 to i-1 do                          // put dummy levels of stack
               with ProfilerStack[j] do begin D0:=-1; timein:= chipset.cycles; end;
        ProfilerStackPointer:= i;
        with ProfilerStack[i] do begin D0:= FullAdr(Chipset.D0); timein:= Chipset.cycles; end;
      end
      else
      Begin
        while ProfilerStackPointer>i do                                    // RemoveTime:= 0;
        Begin
          ProfilerStack[ProfilerStackPointer].timein:=Chipset.cycles-ProfilerStack[ProfilerStackPointer].timein;
          for j:=0 to ProfilerStackPointer-1 do inc(ProfilerStack[j].TimeIn, ProfilerStack[ProfilerStackPointer].timein);
//          RemoveTime:= RemoveTime+ ProfilerStack[ProfilerStackPointer].timein;
          if ProfilerStack[ProfilerStackPointer].d0<>-1 then
          Begin
            j:=0; while (j<ProfilerUsed) and (ProfilerList[j].D0<>ProfilerStack[ProfilerStackPointer].d0) do inc(j);
            if j<ProfilerUsed then
              with ProfilerList[j] do begin inc(nbcalls); inc(time, ProfilerStack[ProfilerStackPointer].timein); end
            else begin
              // bigger list
              if Length(profilerList)<ProfilerUsed+100 then SetLength(profilerList, ProfilerUsed+100);
              ProfilerList[ProfilerUsed].D0:= ProfilerStack[ProfilerStackPointer].D0;
              ProfilerList[ProfilerUsed].time:= ProfilerStack[ProfilerStackPointer].timein;
              ProfilerList[ProfilerUsed].nbcalls:= 1;
              inc(ProfilerUsed);
            end;
          end;
          dec(ProfilerStackPointer);
        end;
//        ProfilerStack[ProfilerStackPointer].timein:= ProfilerStack[ProfilerStackPointer].timein + RemoveTime;
      end;
  end;
  except
  end;

  { now, test the flashpointer stuff if needed }
  if FActiveFlashPtr then
  Begin
    if D0=FLASHPSEC+5 then
    begin
      i:= ReadInteger(Chipset.D1)+5;
      i:= (ReadInteger(i) and $FFF) + ((ReadInteger(i+3) and $ffff) * $1000);
      j:=0; while (j<FlashPtrUsed) and (FlashPtrList[j].value<>i) do inc(j);
      if j<FlashPtrUsed then
        inc(FlashPtrList[j].calls)
      else
      begin
        FlashPtrList[FlashPtrUsed].value:= i; FlashPtrList[FlashPtrUsed].calls:= 1;
        inc(FlashPtrUsed);
      end;
    end;
  end;

  { the real stuff now }
  D02:=FullAdr(D0);    // use it to get the full address
  LastRPLInstr[LastRplPointer]:= D02; // and update the last instruction list.
  inc(LastRplPointer);
  if LastRplPointer>High(LastRPLInstr) then
    LastRplPointer:= Low(LastRPLInstr);
  // we are not in step by step mode and there are breakpoints
  if (RplFullSpeed) and (NbRplBreakPoints<>0) then
  Begin
    // on a breakpoint? if yes, go in step by step mode
    for i:= 0 to NbRplBreakPoints-1 do
      if (RplBreakPoints[i] = D02) or (RplBreakPoints[i] = D0) then
        RplFullSpeed:= False;
  end;
  // if fullspeed or not (fullspeed but in stepover and not on end of the step over)
  if (not RplFullSpeed) and
     ((RplStepOverBreakPoint=-1) or (D0=RplStepOverBreakPoint)) then
  Begin
    RplStepOverBreakPoint:=-1;     // then we are not in stepover anymore
    RplTime:= Chipset.cycles-now;  // compute this instruction time
    PostMessage(FOnRplInstr, DebuggerMessage, 0, 0);  // fire the event
    result:= true;                 // and stop the emulation
  end else
    result:= false;                // continue the emulation
end;

destructor TDebugger.Destroy;
begin
  Deb.Free;
  bl.Free;
  Inherited destroy;
  debugger:= nil;
end;

{ This function allows to open the editor at a specify symbol }
Procedure TDebugger.EditAtSymbol(s: string; doBar:Boolean=False);               // wgg added doBar
var
  i, j: integer;
Begin
  i:= SymbolTable.ValueFromText(s);
  if i<>-1 then
    if Assigned(FCallBackEdit) then // and fire the event
    begin
      if ReadInteger(i) = $11111 then
      begin
        j:= SymbolTable.ValueFromText('_fp_'+s);
        if j=-1 then
        begin
          j:= SymbolTable.ValueFromText(s+'j');
          if j=-1 then
            j:= SymbolTable.ValueFromText('j'+s);
        end;
        if j<>-1 then
          i:= j;
      end;
      FCallBackEdit(Self, i, doBar);                                           // wgg added True to inhibit blue bar
    end;
end;

{ This function allows to open the editor at a specify address }
procedure TDebugger.EditAtAddress(a: integer; doBar:Boolean=False);            // wgg added doBar
begin
  if Assigned(FCallBackEdit) then // and fire the event
    FCallBackEdit(Self, FullAdr(a), doBar);                                    // wgg added False for blue bar
end;

function TDebugger.GetBList(i: integer): TBreakPoint;
begin
  result:= bl[i];
end;

{ fill the Chipset structure if possible }
function TDebugger.GetChipset: TIChipset;
begin
  if deb<>nil then
    deb.GetChipset(FChipset)
  else
    FillChar(FChipset, SizeOf(FChipset), 0);
  result:= FChipset;
end;

// This is called on each ASM instruction (except in non step by step mode).
// The breakpoint process is not handled by the Tdebugger object but by the TLDeb one.
Function TDebugger.AsmInstruction(Sender: TObject): Boolean;
begin
  GetChipset;
  // do not stop if we are within the interupt system and no interupt debugging is checked
  // wgg result:= ZeroReturn or (DebugInt or Not Chipset.Inte);
  result:= true;  // wgg now use EnableSkipInt call
  // fire the event if needed
  if result then    
    PostMessage(FOnAsmInstr, DebuggerMessage, 0, 0);
end;

procedure TDebugger.ListeBreakpoints;
begin
  frmBreakPointChoice.Show;
end;

Function TDebugger.ReadRom(adr: integer): integer;
Begin
  if deb<>nil then
    result:= deb.RomRead(adr)
  else result:= 0;
end;

procedure TDebugger.ReadMemory(var PByte: Array of byte; adress, size: integer);
begin
  if deb<>nil then
    Deb.ReadMemory(PByte, adress and $fffff, size);
end;

// This procedure scan the breakpoint list and
// dispatch the active ones between the
// RPl breakpoints list and the ASM breakpoint engine.
procedure TDebugger.RebuiltBreakPointList;
var
  i: integer;
  asmB: array of integer;
begin
  if deb=nil then exit;
  begin
    // first, build the 2 lists
    NbRplBreakPoints:= 0;
    For i:=0 to BL.Count-1 do
      if TBreakPoint(bl[i]).Active then
        if TBreakPoint(bl[i]).rpl then
        begin
          FChipset:= GetChipset;
          RplBreakPoints[NbRplBreakPoints]:= TBreakPoint(bl[i]).Adresse;
          // if the RPL break points on a program, put the actual break 5 nibbles later.
          if ((RplBreakPoints[NbRplBreakPoints]>$100000) and (Deb.RomRead(RplBreakPoints[NbRplBreakPoints])=DOCOL)) or
             ((RplBreakPoints[NbRplBreakPoints]<$100000) and (ReadInteger(RplBreakPoints[NbRplBreakPoints])=DOCOL)) then
            inc(RplBreakPoints[NbRplBreakPoints], 5);
          inc(NbRplBreakPoints);
        end else begin
          SetLength(asmB, length(AsmB)+1);
          asmB[length(AsmB)-1]:= TBreakPoint(bl[i]).Adresse;
          // wgg if we point at a code prologue, bump break by 5
          // because we will never break on the prologue address
          if ((asmB[length(AsmB)-1]>$100000) and (Deb.RomRead(asmB[length(AsmB)-1])=asmB[length(AsmB)-1]+5)) or
             ((asmB[length(AsmB)-1]<$100000) and (ReadInteger(asmB[length(AsmB)-1])=asmB[length(AsmB)-1]+5)) then
            inc(asmB[length(AsmB)-1], 5);
        end;
  end;
  // send the asm break list to the ASM breakpoint handler
  deb.SetBreakList(asmB);
end;

// remove all breakpoints in the list
procedure TDebugger.RemoveAllBreakPoints;
begin
  While BL.count<>0 do
  Begin
    TObject(bl.Last).Free;
    bl.delete(bl.Count-1);
  end;
  RebuiltBreakPointList;
  if Assigned(OnNewBreak) then
    OnNewBreak(Self);
end;

// delete one of the breakpoints
procedure TDebugger.RemoveBreakPoint(i: integer);
var
  j: integer;
begin
  For j:=0 to bl.Count-1 do
    if TBreakPoint(bl[j]).Adresse=i then
      Begin
        TObject(bl[j]).Free;
        bl.delete(j);
        RebuiltBreakPointList;
        if Assigned(OnNewBreak) then
          OnNewBreak(Self);
        exit;
      end;
end;

// create a new breakpoint
Function TDebugger.SetBreakPoint(Adress: integer; Rpl: Boolean; active: boolean = true): integer;
var
  b: TBreakPoint;
begin
  result:= -1;
  if adress <= 0 then exit;                       // wgg cannot find a good address  
  if not running then exit;
  if BL.Count>1000 then
    raise Exception.Create('Too many Break Points');
  if GetBreakPointIndex(Adress) <> -1 then
  begin
    RemoveBreakPoint(adress);
    Exit;
  end;

  b:= TBreakPoint.Create;
  b.Adresse:= Adress;
  b.Rpl:= Rpl;
  b.Active:= active;
  result:= BL.Add(b);
  RebuiltBreakPointList;
  if Assigned(OnNewBreak) then
    OnNewBreak(Self);
end;

// wgg check to see if we have a valid breakpoint
function TDebugger.GetBreakPointIndex(Adress: integer): integer;                    // wgg
var
  i: integer;
begin
  result:= -1;
  if adress <= 0 then exit;                                   // wgg cannot find a good address  
  if not running then exit;
  for i:= 0 to bl.Count-1 do
    if TBreakPoint(bl[i]).Adresse=adress then
    begin
      result:= i;
      Exit;
    end;
end;

// wgg set file index and lines for each bp
procedure TDebugger.SaveBPSource;
var
  i, j: integer;
  SaveP: PPos;
begin
  setLength(bsl, 1000);                             // wgg make enough room for all
  j:= 0;
  for i:=0 to bl.Count-1 do
    with TBreakPoint(bl[i]) do
    begin
      SaveP:= SourceFromAdr(Adresse);
      if SaveP<>nil then
      Begin
        bsl[j].fileName:= SaveP^.FileName;         // wgg need to copy data, may not exist after build
        bsl[j].line:= SaveP^.line;
        bsl[j].IsRPL:= Active;                     // wgg misuse a field for convenience!
        inc(j);
      end
      else fil:= Nil;
    end;
  setLength(bsl, j);
end;

// wgg set BP in files using saved file index and lines
procedure TDebugger.recomputeBP;
var
  i, j: integer;
  RPL: boolean;
begin
  RemoveAllBreakPoints;
  for i:= 0 to high(bsl) do
    with TPos(bsl[i]) do
    try
       fil:= GetFileFromFileName(fileName);
       j:= fil.AdrFromLine(TFileName(fileName), line, RPL);                     // wgg RPL is returned
      // ** IsRPL has been used to store breakpoint's Active status!!
      if j > 0 then  DEbugger.SetBreakPoint(j, RPL, isRPL);  
    except
    end;
  setLength(bsl, 0);
  frmBreakPointChoice.FillListBox;
  RebuiltBreakPointList;
end;

// fire an event if a debugger is stopped or started
procedure TDebugger.SetOnStateChange(const Value: TNotifyEvent);
begin
  if assigned(FOnStateChange) then
    FOnStateChange(self);
  FOnStateChange := Value;
  if assigned(FOnStateChange) then
    FOnStateChange(self);
end;

procedure TDebugger.WriteMemory(var PByte: Array of byte; adress, size: integer);
begin
  if deb<>nil then
    Deb.WriteMemory(PByte, adress and $fffff, size);
end;

// step over an ASM instruction
// if the next instruction is not a GOSUB, perform a simple STEPIN
procedure TDebugger.StepOverAsm;
var
  b: array [0..1] of byte;
begin
  if deb=nil then exit;
  if not deb.AsmDebug then exit;
  ReadMemory(b, Chipset.PC, 2);
  if (b[0]=7) then  // GOSUB?
    Deb.StepOver(Chipset.PC+4)
  else
    if b[0]=8 then
       if b[1]=$E then // GOSUBL
         Deb.StepOver(Chipset.PC+6)
       else
        if b[1]=$F then // GOSBVL?
          Deb.StepOver(Chipset.PC+7)
        else
          Deb.StepIn
    else
      Deb.StepIn
end;

procedure TDebugger.StepOverRpl;
begin
  if Running and (not deb.AsmDebug) then
  Begin
    now:= Chipset.cycles;
    if (NextRplStepOver <= 0) then
    Begin
      RplStepOverBreakPoint:= -1;
      StepIn;
    end else
      RplStepOverBreakPoint:= NextRplStepOver;
    Continue;
  end;
end;

// memory read procedures
function TDebugger.ReadInteger(Adr: integer): integer;
Var
  buf: Array[0..4] Of byte;
Begin
  ReadMemory(buf, adr, 5);
  result := buf[0] + buf[1] * 16 + buf[2] * 256 + buf[3] * 16 * 256 + buf[4] * 65536;
End;

function TDebugger.ReadInt64(Adr: Integer): int64;
Var
  buf: Array[0..15] Of byte;
  v1, v2: int64;
begin
  ReadMemory(buf, adr, 16);
  v1 := buf[0] +       buf[1]shl 4 +   buf[2]shl 8 +   buf[3]shl 12 +
        buf[4]shl 16+  buf[5]shl 20 +  buf[6]shl 24 +  buf[7]shl 28;
  v2 := buf[8] +       buf[9]shl 4 +   buf[10]shl 8 +  buf[11]shl 12 +
        buf[12]shl 16+ buf[13]shl 20 + buf[14]shl 24 + buf[15]shl 28;
  result:= v1+(v2 shl 32);
end;

{
Cyril's Comment
This expand an address in the current view in a fully address
Decoding is based on the 6th nibble. if it's 0, it means that it's directly
a saturn address. if it's non 0, 6th nibble is bank number, 5 other nibbles
are suposed to be < 40000 and indicate a point in the bank. Now, bank 0's
number is... 0 and that causes a clash....

wgg I added 1,000,000 for any FLASH address so ow a 0 bank works!
}

function TDebugger.FullAdr(adr: integer): integer;
begin
  if CalcModel='X'        // wgg HP49
  then
    try
      if adr>=$100000 then                     // wgg we will make it FLASH
        adr:= (adr Or $1000000) and $1F3FFFF;  // wgg 1,000,000 means flash
      if adr>=$80000 then // in ram? do nothing
        result:= adr
      else
        if adr<$40000 then  // in view1?                      // wgg this does not work for HP48
          result:= adr+(chipset.RomV1*$100000)+$1000000       // wgg add FLASH bit
        else                // or view2?
          result:= (adr and $3ffff)+(chipset.RomV2*$100000)+$1000000;       // wgg add FLASH bit
    except
      result:= adr;
    end
  else
    result:= adr;
end;

function TDebugger.GetRunning: Boolean;
begin
  result:= Deb<>nil;
end;

// change the debugger
procedure TDebugger.SetDebugType(const Value: TDebType);
var
  f, fPort2: string;
begin
  if DebugType = Value then exit;
  if DebugType <> DtNone then Deb.EmDestroy; // wgg kill debugger, see Emu49Close for cleanup
  if DebugType = DtNone then                 // now that its gone, open a new one
    begin
      if FPort2File<>'' then fPort2:= ExpandFileName(FPort2File);   // wgg
      f:= ExpandFileName(FRamFile);          // wgg
      if Not FileExists(f) then
      begin
        Application.MessageBox(PChar('Emu RAM file specified, "'+f
                         +','+#10#13'does not exist, emulation canceled.'),'Warning');
        exit;
      end;
      RplFullSpeed:= True;
      RplStepOverBreakPoint:= -1;
      if Value=DtEMU49 then
      begin
        Deb:= TEmu49Deb.Create(AsmInstruction, RplInstruction,
                DebBye, f, fPort2);       // wgg Ram file will make the difference 48 or 49
        if Deb.CalcModel='^' then If Deb.EmDestroy then Deb.Free;      // wgg we aborted for some sort of file reason
      end;
      DebugInt:= DebugInt;                                             // wgg will copy state to EMU on write
      if Assigned(FOnStateChange) then
        FOnStateChange(self);
    end;
end;

function TDebugger.GetDebugType: TDebType;
begin
  if Deb=nil then
    result:= DtNone
  else
    result:= Deb.DType;
end;

// call back when the debugger stops
procedure TDebugger.DebBye(Sender: TObject);
begin
  Deb:= Nil;
  if Assigned(FOnStateChange) then
    FOnStateChange(self);
end;

// emulation control
procedure TDebugger.continueAsm;
begin
  if (deb<>nil) and (deb.AsmDebug) then
    Deb.Run;
end;

procedure TDebugger.continueRpl;
begin
  if not running then exit;
  if deb.AsmDebug then exit;
  RplStepOverBreakPoint:= -1;
  RplFullSpeed:= True;
  if Deb<>nil then
    Deb.Run;
  frmDebugRPL.ClearAll;
end;

function TDebugger.InAsmStep: boolean;       // wgg added for ASM Debug window
begin
  if (Deb<>nil) and (deb.AsmDebug) then result:= true
  else result:= false;
end;

function TDebugger.InRplStep: boolean;       // wgg added to block build/load
begin
  if (Deb<>nil) and (Not RplFullSpeed) then result:= true
  else result:= false;
end;

procedure TDebugger.StepInAsm;
begin
  if (Deb<>nil) and (deb.AsmDebug) then
    Deb.StepIn;
end;

procedure TDebugger.StepInRpl;
begin
  if Running and (not deb.AsmDebug) then
  begin
    now:= Chipset.cycles;
    Deb.run;
  end;
end;

procedure TDebugger.SaveChipset;
begin
  if deb<>nil then
    deb.SetChipset(FChipset);
end;

procedure TDebugger.StopAsm;
begin
  if deb<>nil then
    deb.stop;
end;

procedure TDebugger.StopRpl;
begin
  RplStepOverBreakPoint:= -1;
  RplFullSpeed:= false;
end;

function TDebugger.GLastInstr: TLinstr;
begin
  result:= Deb.LastAsmInstr;
end;

procedure TDebugger.SetEraseStk(const Value: boolean);
begin
  FEraseStk := Value;
end;

function TDebugger.GLastRplInstr: TLinstr;
var
  i, j: integer;
begin
  j:= LastRplPointer+High(LastRPLInstr);
  for i:= High(LastRPLInstr) downto Low(LastRPLInstr) do
  Begin
    j:= j-1;
    result[i]:= LastRPLInstr[j and NbLastInstr];
  end;
end;

procedure TDebugger.WriteInteger(adr, value: integer);
Var
  buf: Array[0..4] Of byte;
  i: integer;
Begin
  for i:= 0 to 4 do
  Begin;
    buf[i]:= value and $f;
    value:= value div 16;
  end;
  WriteMemory(buf, adr, 5);
End;

procedure TDebugger.LoadFile(f: TFileName);
begin
  if deb<>nil then
    deb.LoadFile(f);
end;

procedure TDebugger.StepIn;
begin
  if not running then exit;
  if deb.AsmDebug then
    StepInAsm
  else
    if (not RplFullSpeed) then
      if RplStepOverBreakPoint=-1 then
        StepInRpl;
end;

procedure TDebugger.StepOver;
begin
  if not running then exit;
  if deb.AsmDebug then
    StepOverAsm
  else
    if (not RplFullSpeed) then
      if RplStepOverBreakPoint=-1 then
        StepOverRpl;
end;

procedure TDebugger.ExecKey(s: string; action: TThreadMethod);
begin
  if deb<>nil then
    deb.execKey(s, action);
end;

// wgg wait for debugger ShutDn signal or timeout
// returns true if got the signal, false for timeout
function TDebugger.WaitShutDn: boolean;
begin
  if deb<>nil then result:= deb.WaitShutDn
  else result:= false;
end;

// wgg wait for debugger busy flag off or timeout
// returns true if got the signal, false for timeout
function TDebugger.WaitBusy: boolean;
begin
  if deb<>nil then result:= deb.WaitBusy
  else result:= false;
end;

procedure TDebugger.SetDebugInt(const Value: Boolean);
begin
  FDebugInt := Value;
  if assigned(deb) then
    Deb.DebugInterrupt(Value);
end;

function TDebugger.crntAdrs: integer;
begin
  result:= deb.LastAsmInstr[253]
end;

{ TBreakPoint }

constructor TBreakPoint.Create;
begin
  inherited create;
  rpl:= false;
  Active:= True;
end;

{ TLDeb }

constructor TLDeb.Create(AsmInstr: TNotifyFunction; RplInstr: TNotifyRplFunction;
            bye: TNotifyEvent; Ram: TFileName; Port2: TFileName);          // wgg
begin
  inherited create;
  FAsmInstr:= AsmInstr;
  FRplInstr:= RplInstr;
  FBye:= Bye;
end;

procedure TLDeb.DebugInterrupt(value: boolean);
begin

end;

{
The first boot!
At the beginning of the project, we were working on the emulator and one day came
the first prototype! We downloaded the software in the prototype and . It did not boot.
How could we know what was failing?
Were we going for blind debugging?
No way!
We review the concept of 'Printf' and added everywhere in the boot code a 'SendChar'
call sending a different char to the serial port and monitoring it on a terminal.
It was failing between the 'P' and the 'Q'!
This means it was failing between the CONFIG and the main kernel boot.
The CONFIG section, is the moment where the config object of each library is executed.
And it's an RPL program, no way to debug it like an ASM program with our 'printf'.
We had to find something else. It was the Serial RPL Debugger!
The Serial RPL debugger was a set of two things:
- A modification of the debugger allowing it to debug throught the serial port
rather that from an emulator.
- A little ASM code allowing the PC to read the HP memory and registers.
The LOOP macro of  RPLCOMP had been modified in order to be a call to the serial
debugger rather that a A=DAT0.A D0+5 PC=(A), allowing to let the PC speak with
the calculator after each RPL instruction.
One of the features of the Serial debugger was a short 'beep' at each call of
the serial code, allowing us to count the RPL instructions executed. Around 3 per second!
Tick, Tick, Tick, ..
The calculator start booting, but It was awfully slow, the old 'CATROMPID' function
was so slow! After a brief computation, we discover that building the list of the
present libraries was a 2 hours operation, sSo Jean-Yves rewrote it in plain ASM
and after a new system download, we started again!
This allowed us to find the bug (a relay nasty one!) and for the first time
Tick, Tick, Tick, . For the first time, a HP49G was booting! Very slowly (try
to imagine, 3 RPL instruction per second)! It was fully functional but way to slow!
So we removed the call to the Serial debugger, recompiled and started again
downloaded it on a prototype and start playing with it (at full speed this time)!
And then, suddenly, Tick, Tick, Tick,  The first prototype started Ticking again!
After 10 minutes of inactivity, the calculator was turning itself off!
This was the story of the first HP49G boot.


PS1: The Serial debugger was on of the TLDeb childs but It was not improved when new property
had been added to the bas class and it disapears. Peace on it's soul.

PS2: The nasty bug was comming from the fact that the YORKE is always reading one more
nibble when reading from the memory, and they where one place when a WRITE was use instead of
a read in order to switch the flash, and therfore, it was not swaping property (It was JYA fault:-)
}

{ TEmu49Deb }

function TEmu49Deb.AsmDebug: boolean;
begin
  result:= StepByStep;
end;

procedure TEmu49Deb.AsmInstruction(sender: tobject);
begin
  EmuClearBreakpoint(AsmStepOverBreakPoint, EMU_BP_EXEC);
  StepByStep:= True;
  while not FAsmInstr(Self) do
    EmuStep;
  Debugger.ZeroReturn:= False;
end;

var
  Emu49Deb: TEmu49Deb;

procedure Emu49Close(); stdcall;
begin
  Emu49deb.AutoClose:= true;
  Debugger.RemoveAllBreakPoints;               // wgg
  Debugger.RplFullSpeed:= True;                // wgg inhibits builds
  frmDebugRPL.ClearAll;                        // wgg
  frmSourceEdit.ClearAdrsMark;               // wgg clear pc marker
  frmDebugEMU.ActionContinueExecute(Nil);           // wgg if deb=nil, this is ok
  frmProject.FileAdr.Text:='';
  Emu49Deb.free;
end;

// wgg added this to interpret new calc types from EMU48
// for the 48Gii and 49G+.  Treat these all as X (or 49G)
//               '6' = HP38G with 64KB RAM
//               'A' = HP38G
//               'E' = HP39/40G
//               'S' = HP48SX
//               'G' = HP48GX
//               'X' = HP49G
//               'P' = HP39G+     // JYA changes
//               'Q' = HP49G+
//               '2' = HP48GII
function LimitedEmuCalculatorType(): char;
var
  t:char;
begin
    t:= EmuCalculatorType;
    case t of
    'X': Result:= 'X';
    'P': Result:= 'X';
    'Q': Result:= 'X';
    '2': Result:= 'X';
    else Result:= t;
    end;
end;

procedure breakpoint(nBreakType: integer); stdcall;  // wgg integer was longbool
begin
  case nBreakType of
  BN_ASM:    Emu49Deb.AsmInstruction(Emu49Deb);     // ASM breakpoint
  BN_RPL:    Emu49Deb.RplInstruction(Emu49Deb);     // RPL breakpoint
  BN_ASM_BT: Emu49Deb.AsmInstruction(Emu49Deb);     // could be either one, entry to code module
  end;
end;

constructor TEmu49Deb.Create(AsmInstr: TNotifyFunction; RplInstr: TNotifyRplFunction;
           bye: TNotifyEvent; Ram: TFileName; Port2: TFileName);
var
  b: boolean;
begin
  inherited create(AsmInstr, RplInstr, bye, ram, port2);
  AutoClose:= false;
  Emu49Deb:= self;
  if Port2='' then                              // wgg if pass '' to EmuCreatEx, get no valid port2
    b:= EmuCreate(pchar(ram))                   // wgg returns True if it failed!
  else
    b:= EmuCreateEx(pchar(ram), pchar(Port2));  // wgg returns True if it failed!
  if Not b then
  begin
    AutoClose:= true;
    EmuCallBackClose(Emu49Close);
    EmuCallBackDebugNotify(breakpoint);
    EmuInitLastInstr(NbLastInstr+1, @(lasti));
    EmuEnableRplBreakpoint(true);
    AsmStepOverBreakPoint:= 0;
    StepByStep:= False;
    CalcModel:= LimitedEmuCalculatorType;
  end else
    CalcModel:='^';   // wgg signal that we failed
end;

procedure TEmu49Deb.DebugInterrupt(value: boolean);
begin
  EmuEnableSkipInterruptCode((not value));
end;

destructor TEmu49Deb.destroy;
begin
  if not AutoClose then
  begin
    EmuStop;
    EmuDestroy;
  end;
  Fbye(self);
  inherited destroy;
end;

function TEmu49Deb.DType: TDebType;
begin
  result:= DtEmu49;
end;

type
  TKeyPressThread = class (tthread)
    s: string;
    action: TThreadMethod;
    constructor create(s:string; action: TThreadMethod);
    procedure Execute; override;
  end;

// wgg wait for debugger ShutDn signal or timeout
// returns true if got the signal, false for timeout
function TEmu49Deb.WaitShutDn: boolean;
var
  i: integer;
const
  maxTime: integer = 4000;        // maximum time for waiting
  oneTime: integer = 100;          // time for each sleep
begin
  i:= 0;
  result:= true;
  while ((EmuGetRegister(EMU_REGISTER_FLAGS) and $8000)shr 15) = 0 do
    if i > maxTime then
    begin
      result:= false;
      break;
    end else begin
      inc(i, oneTime);
      Application.ProcessMessages;
      sleep(oneTime);
    end;
  if Not Result then result:= WaitBusy;   // wgg also make sure no busy Annunciator  
end;

// wgg wait for debugger Busy flag off or timeout
// returns true if got the signal, false for timeout
// This did not help the Loading to EMU process
function TEmu49Deb.WaitBusy: boolean;
var
  i: integer;
  flags: byte;
const
  oneTime: integer = 100;          // time for each sleep

  function Wait(bit: integer; maxTime: integer): boolean;
  begin
  i:= 0;
  result:= true;
  repeat
    if i > maxTime then
    begin
      result:= false;
      break;
    end else begin
      inc(i, oneTime);
      Application.ProcessMessages;
      sleep(oneTime);
      Debugger.ReadMemory(flags, $10C, 1);
    end;
  until (flags and 1) = bit;
  end;

begin
//  Wait(1, 1000);                  // wait short time for it to go positive
  Result:= wait(0, 3500);        // wait for it to clear
end;

procedure TEmu49Deb.ExecKey(s: string; action: TThreadMethod);
begin
  TKeyPressThread.Create(s, action);
end;

constructor TKeyPressThread.create(s: string; action: TThreadMethod);
begin
  inherited create(true);
  self.action:= action;
  self.s:= s;
  freeOnterminate:= true;
  resume;
end;

procedure TKeyPressThread.execute;
const
//
// Codes for HP49
//
// wgg: There are program calls for #17 (a red shift), #27 (ON) and #19 (after autoload)
//      These values do not make any sense that I can see and they conflict with published key tables
//
//      I make the following definitions:
//      #01=del,  #02=bksp,  #9=alpha,  #13=ENTER,  #16=left-shift,  #19=right-shift,  #27=ON
//
   outv49: array [1..47] of integer = (5,  5,  5,  5,  5,  5,  5,  4,  3,  2,  1,  0,  4,  3,  2,  1,  4,  3,  2,  1,  0,  4,  3,  2,  1,  0,  7,  3,  2,  1,  0,  7,  3,  2,  1,  0,  7,  3,  2,  1,  0,  0,  3,  2,  1,  0,  0);
    inv49: array [1..47] of integer = (0,  1,  2,  3,  4,  5,  7,  7,  7,  7,  7,  7,  6,  6,  6,  6,  5,  5,  5,  5,  5,  4,  4,  4,  4,  4,  3,  3,  3,  3,  3,  2,  2,  2,  2,  2,  1,  1,  1,  1,  1, 15,  0,  0,  0,  0,  6);
   keyv49=                            'A'+'B'+'C'+'D'+'E'+'F'+'G'+'H'+'I'+'J'+'K'+'L'+'M'+'N'+'O'+'P'+'Q'+'R'+'S'+'T'+'U'+'V'+'W'+'X'+'Y'+'Z'+#9+ '7'+'8'+'9'+'*'+#16+'4'+'5'+'6'+'-'+#19+'1'+'2'+'3'+'+'+#27+'0'+'.'+' '+#13+#02; // last two are ENTER BKSP

//
//    wgg - Codes for HP-48
//
   outv48: array [1..49] of integer = (1,  8,  8,  8,  8,  8,  2,  7,  7,  7,  7,  7,  0,  6,  6,  6,  6,  6,  3,  5,  5,  5,  5,  5,  4,  4,  3,  3,  3,  3,  3,  2,  2,  2,  2,  2,  1,  1,  1,  1,  1,  0,  0,  0,  0,  0,  4,  4,  4);
    inv48: array [1..49] of integer = (4,  4,  3,  2,  1,  0,  4,  4,  3,  2,  1,  0,  4,  4,  3,  2,  1,  0,  4,  4,  3,  2,  1,  0,  3,  2,  5,  3,  2,  1,  0,  5,  3,  2,  1,  0,  5,  3,  2,  1,  0, 15,  3,  2,  1,  0,  4,  1,  0);
   keyv48=                            'A'+'B'+'C'+'D'+'E'+'F'+'G'+'H'+'I'+'J'+'K'+'L'+'M'+'N'+'O'+'P'+'Q'+'R'+'S'+'T'+'U'+'V'+'W'+'X'+'Y'+'Z'+#9+ '7'+'8'+'9'+'/'+#16+'4'+'5'+'6'+'*'+#19+'1'+'2'+'3'+'-'+#27+'0'+'.'+' '+'+'+#13+#01+#02; // last three are ENTER DEL bkspc
var
  i: integer;
  c: char;
begin
//  sleep(500);
  c:= #0;
  for i:= 1 to length(s) do
  begin
    if s[i]=c then sleep(250); // avoid problem when the same key is press twice in a row
    c:= s[i];
    if LimitedEmuCalculatorType ='X' then                        // wgg - its HP49
    begin
      EmuSimulateKey(true, outv49[pos(s[i], keyv49)], 1 shl  inv49[pos(s[i], keyv49)]);  Sleep(100);
      EmuSimulateKey(False, outv49[pos(s[i], keyv49)], 1 shl  inv49[pos(s[i], keyv49)]); Sleep(50);
    end else                                     // wgg - else its a HP48
    begin
      EmuSimulateKey(true, outv48[pos(s[i], keyv48)], 1 shl  inv48[pos(s[i], keyv48)]);  Sleep(100);
      EmuSimulateKey(False, outv48[pos(s[i], keyv48)], 1 shl  inv48[pos(s[i], keyv48)]); Sleep(50);
    end;
  end;
  if assigned(action) then
  Begin
    sleep(600);
    Synchronize(action);
  end;
end;

procedure TEmu49Deb.GetChipset(var Chipset: TIChipset);
var
  i, V1: integer;
  function gr32(i: integer): int64;
  begin
    result:= EmuGetRegister(i+1);
    result:= result shl 32;
    result:= result+EmuGetRegister(i);
  end;
begin
  chipset.A       := gr32(EMU_REGISTER_AL);
  chipset.B       := gr32(EMU_REGISTER_BL);
  chipset.C       := gr32(EMU_REGISTER_CL);
  chipset.D       := gr32(EMU_REGISTER_DL);
  chipset.R0      := gr32(EMU_REGISTER_R0L);
  chipset.R1      := gr32(EMU_REGISTER_R1L);
  chipset.R2      := gr32(EMU_REGISTER_R2L);
  chipset.R3      := gr32(EMU_REGISTER_R3L);
  chipset.R4      := gr32(EMU_REGISTER_R4L);
  chipset.D0      := EmuGetRegister(EMU_REGISTER_D0);
  chipset.D1      := EmuGetRegister(EMU_REGISTER_D1);
  chipset.PC      := EmuGetRegister(EMU_REGISTER_PC);
  chipset.P       := EmuGetRegister(EMU_REGISTER_FLAGS) and $f;
  chipset.carry   := EmuGetRegister(EMU_REGISTER_FLAGS) and $200 <> 0;
  chipset.Inte    := Not (EmuGetRegister(EMU_REGISTER_FLAGS) and $400 <> 0);
  chipset.Decimal := EmuGetRegister(EMU_REGISTER_FLAGS) and $100 <> 0;
  chipset.RomV1   := EmuGetRegister(EMU_REGISTER_VIEW1);
  chipset.RomV2   := EmuGetRegister(EMU_REGISTER_VIEW2);

  if CalcType = 'G' then       // wgg HP48
  begin
    V1:= (chipset.RomV1 And 2) shr 1;
    chipset.RomV2:= (chipset.RomV2 And $F) Or ((chipset.RomV1 And 1) shl 4);
    chipset.RomV1:= V1;
  end;

  Chipset.cycles  := EmuGetRegister(EMU_REGISTER_CLKL)+EmuGetRegister(EMU_REGISTER_CLKH) shl 32;
  chipset.CRC     := EmuGetRegister(EMU_REGISTER_CRC);                   // wgg
  for i:= 0 to 15 do
    chipset.ST[i] := not((EmuGetRegister(EMU_REGISTER_FLAGS) shr 16) and (1 shl i) = 0);
  for i:= 0 to 3 do
    chipset.HST[i]:= not((EmuGetRegister(EMU_REGISTER_FLAGS) shr 4) and (1 shl i) = 0);
  for i:= 0 to 7 do
    Chipset.RSTK[i]:= EmuGetRegister((integer(EmuGetRegister(EMU_REGISTER_RSTKP))+i+1) and $7 + EMU_REGISTER_RSTK0);
end;

function TEmu49Deb.LastAsmInstr: TLinstr;
var
  i: Dword;
begin
  EmuGetLastInstr(i);
  result:= lasti;
end;

function TEmu49Deb.EmDestroy: boolean;
begin
  result:= EmuDestroy;
end;

procedure TEmu49Deb.LoadFile(f: TFileName);
begin
  EmuCallBackDebugNotify(Nil);             // wgg EMU will call our DebugNotify during a load
  WaitShutDn;                              // wgg make sure EMU is ready for it
  EmuLoadObject(Pchar(f));                 // because EMU turns debug off then on for loading!
  EmuCallBackDebugNotify(breakpoint);      // we will hang due to single thread on out side if this happens
end;

procedure TEmu49Deb.ReadMemory(var PByte: array of byte; adress, size: integer);
var
  i: integer;
begin
  for i:=0 to size-1 do
    EmuGetMem(adress+i, PByte[i]);
end;

function TEmu49Deb.RomRead(a: integer): integer;
var
  b: PByte;
  size: DWORD;
begin
  result:= 0;
  b:= EmuGetRom(size);
  a:= (a and $3FFFF) + (a and $F00000 div 4);
  inc(b, a);
  if a+4<=integer(size) then
    for size:= 0 to 4 do
    Begin
      result:= (result shr 4) or (b^ shl 16);
      inc(b);
    end;
End;

procedure TEmu49Deb.RplInstruction(sender: tobject);
var
  s: string[3];
  pc: integer;
begin
  if StepByStep then
    AsmInstruction(self)
  else begin
    SetLength(s, 20);
    pc:= LastAsmInstr[253];
    ReadMemory(PByte(@s[1])^, pc and $FFFFF, 3);
    if (copy(s,1,3)<>#1#6#4) or
       (not FRplInstr(self, EmuGetRegister(EMU_REGISTER_D0)-5)) then
      EmuRun;
  end;
end;

procedure TEmu49Deb.Run;
begin
  StepByStep:= False;
  EmuRun;
end;

procedure TEmu49Deb.SetBreakList(b: array of integer);
var
  i: integer;
begin
  for i:=0 to high(blist) do
    if blist[i] and $1F00000 = 0 then                 // wgg bug was b[i], added $1,000,000 for FLASH
      EmuClearBreakpoint(blist[i], EMU_BP_EXEC)
    else
      EmuClearBreakpoint((blist[i] and $3ffff) or ((blist[i] and $f00000) shr 2), EMU_BP_ROM);
  SetLength(blist, length(b));
  for i:=0 to high(blist) do
  Begin
    blist[i]:= b[i];
    if b[i] and $1F00000 = 0 then                 // wgg added $1,000,000 for FLASH
      EmuSetBreakpoint(blist[i], EMU_BP_EXEC)
    else   
      EmuSetBreakpoint((blist[i] and $3ffff) or ((blist[i] and $f00000) shr 2), EMU_BP_ROM);
  end;
end;

procedure TEmu49Deb.SetChipset(Chipset: TIChipset);
var
  i: integer;
  j: DWord;

begin
  EmuSetRegister(EMU_REGISTER_AL, chipset.A and $ffffffff);
  EmuSetRegister(EMU_REGISTER_BL, chipset.B and $ffffffff);
  EmuSetRegister(EMU_REGISTER_CL, chipset.C and $ffffffff);
  EmuSetRegister(EMU_REGISTER_DL, chipset.D and $ffffffff);
  EmuSetRegister(EMU_REGISTER_R0L, chipset.R0 and $ffffffff);
  EmuSetRegister(EMU_REGISTER_R1L, chipset.R1 and $ffffffff);
  EmuSetRegister(EMU_REGISTER_R2L, chipset.R2 and $ffffffff);
  EmuSetRegister(EMU_REGISTER_R3L, chipset.R3 and $ffffffff);
  EmuSetRegister(EMU_REGISTER_R4L, chipset.R4 and $ffffffff);
  EmuSetRegister(EMU_REGISTER_AH, (chipset.A shr 32) and $ffffffff);
  EmuSetRegister(EMU_REGISTER_BH, (chipset.B shr 32) and $ffffffff);
  EmuSetRegister(EMU_REGISTER_CH, (chipset.C shr 32) and $ffffffff);
  EmuSetRegister(EMU_REGISTER_DH, (chipset.D shr 32) and $ffffffff);
  EmuSetRegister(EMU_REGISTER_R0H, (chipset.R0 shr 32) and $ffffffff);
  EmuSetRegister(EMU_REGISTER_R1H, (chipset.R1 shr 32) and $ffffffff);
  EmuSetRegister(EMU_REGISTER_R2H, (chipset.R2 shr 32) and $ffffffff);
  EmuSetRegister(EMU_REGISTER_R3H, (chipset.R3 shr 32) and $ffffffff);
  EmuSetRegister(EMU_REGISTER_R4H, (chipset.R4 shr 32) and $ffffffff);
  EmuSetRegister(EMU_REGISTER_D0, chipset.D0);
  EmuSetRegister(EMU_REGISTER_D1, chipset.D1);
  EmuSetRegister(EMU_REGISTER_PC, chipset.PC);

  if CalcType = 'G' then       // wgg HP48
  begin
    EmuSetRegister(EMU_REGISTER_VIEW1, ((chipset.RomV1 And $1) shl 1) Or ((chipset.RomV2 shr 4) And $1));
    EmuSetRegister(EMU_REGISTER_VIEW2, chipset.RomV2 And $F);
  end else begin
    EmuSetRegister(EMU_REGISTER_VIEW1, chipset.RomV1);
    EmuSetRegister(EMU_REGISTER_VIEW2, chipset.RomV2);
  end;

  EmuSetRegister(EMU_REGISTER_CRC, chipset.CRC);                        // wgg

  j:= 0;
  for i:= 0 to 15 do
    if chipset.st[i] then
      j:= j + (1 shl i);
  j:= j shl 12;
  for i:= 0 to 3 do
    if chipset.hst[i] then
      j:= j or (1 shl i);
  j:= j shl 4;
  j:= j+DWord(Chipset.p and $f);
  if chipset.carry then j:= j + $200;
  if Not chipset.inte then j:= j + $400;
  if chipset.decimal then j:= j + $100;
  j:= j+ EmuGetRegister(EMU_REGISTER_FLAGS) and $800;
  EmuSetRegister(EMU_REGISTER_FLAGS, j);
  for i:= 0 to 7 do
    EmuSetRegister((integer(EmuGetRegister(EMU_REGISTER_RSTKP))+i+1) and $7 + EMU_REGISTER_RSTK0,
                Chipset.RSTK[i]);
end;

procedure TEmu49Deb.StepIn;
begin
  if StepByStep then
    EmuStep;
end;

procedure TEmu49Deb.StepOver(adr: integer);
begin
  if StepByStep then
  Begin
    StepByStep:= False;
    EmuSetBreakpoint(adr, EMU_BP_EXEC);
    AsmStepOverBreakPoint:= Adr;
    EmuRun;
  end;
end;

procedure TEmu49Deb.Stop;
begin
  StepByStep:= True;
  EmuStop;
end;

procedure TEmu49Deb.Switch(adr: integer);
begin
  EmuSetRegister(EMU_REGISTER_VIEW1, adr div 32);
  EmuSetRegister(EMU_REGISTER_VIEW1, adr div 2);
end;

procedure TEmu49Deb.WriteMemory(var PByte: array of byte; adress, size: integer);
var
  i: integer;
begin
  for i:=0 to size-1 do
    EmuSetMem(adress, PByte[i]);
end;

initialization
  Debugger:= TDebugger.Create(application);;

finalization
  Debugger.Free;
end.
