{*******************************************************}
{                                                       }
{       Aerodynamica Components                         }
{       SynAutoCorrect 2.x                              }
{                                                       }
{       Copyright (c) 1996-2003, Aerodynamica Software  }
{                                                       }
{       Author: Aaron Chan                              }
{       Portions by Greg Nixon and Jan Fiala            }
{                                                       }
{*******************************************************}

{
  @author: Aaron Chan
  @url: http://aerodynamica.idz.net
  @comp-url: http://www.freeasphost.co.uk/asw/products.asp?id=SynAC_2
  @email: aerodynamica@idz.net
  @last-updated: 30/09/02
  @documentation: See SynAutoCorrect.html.
  @history:

    ! comment     * changed     + added
    - removed     @ bug-fixed   # todo

    30/09 - Release 2.1:
      @ Fixed bug when user KeyDown and MouseDown events weren't fired.  
      + Added INI_FILES and REGISTRY compiler defines (to minimize code size
        if you don't need these features).  
      * Further tidy-up of code.
      * Quite a few minor bug-fixes and tweaks.
      * Items editor enhanced.
      * Improved demo.             
      * Registry and INI file entries are saved in a new and improved way, which
        overcomes some limitations set by the old method. If you still want to
        use the old method, define OLD_SAVE_METHOD.

    31/07 - Revision 2.01:
      @ Fixed bug which occured when undefining SOUND_SUPPORT (reported by
        Stefan Ascher).

    30/07 - First public release of version 2.0:
      @ MANY bugs fixed and small tweaks everywhere in the code (some
        courtesy of Jan Fiala).
      + Ability to play an optional WAVE file (or beep) on correction.
      + Options set.
      * New demo.
}

unit SynAutoCorrect;

{$I SynEdit.inc}

{.$DEFINE OLD_SAVE_METHOD}                       // Use the old saving method
{$DEFINE INI_FILES}                              // Enable INI files support.

{$IFDEF SYN_CLX}
{$ELSE}      
  {$DEFINE REGISTRY}                             // Enable registry support.
  {$DEFINE SOUND_SUPPORT}                        // Enable sound support.
{$ENDIF}

{$IFDEF INI_FILES}
  {$DEFINE REGISTRY_OR_INI}
{$ENDIF}
{$IFDEF REGISTRY}
  {$DEFINE REGISTRY_OR_INI}
{$ENDIF}

interface

uses
  Classes,
  SysUtils,

{$IFDEF SYN_CLX}
  QGraphics,
  QControls,
  QForms,
  QDialogs,
  Types,
{$ELSE}
  Windows,
  Messages,
  Graphics,
  Controls,
  Forms,
  Dialogs,
{$ENDIF}

{$IFDEF SOUND_SUPPORT}
  MMSystem,
{$ENDIF}

{$IFDEF INI_FILES}
  IniFiles,
{$ENDIF}

{$IFDEF REGISTRY}
  Registry,
{$ENDIF}

  SynEditMiscProcs,
  SynEditTypes,
  SynEditHighlighter,
  SynEdit;
      
{$IFDEF OLD_SAVE_METHOD}
const
  FStrSepChar = '|';
{$ENDIF}

type
  { Internal "hack" class :) }
  TAsCustomSynEdit = class(TCustomSynEdit);

  { Used to store KeyDown and MouseDown events. Anyone know of a better way to
    hook into editor events? }
  TEditorEventsRec = record
    kd: TKeyEvent;
    md: TMouseEvent;
  end;
  PEditorEventsRec = ^TEditorEventsRec;

{$IFDEF SOUND_SUPPORT}
  TAsSoundOption = (assoAsync, assoNoDefault, assoNoWait);
  TAsSoundOptions = set of TAsSoundOption;
  TAsSoundType = (asstAlias, asstFileName, asstBeep);

  TAsSoundInfo = class(TPersistent)
  private
    FFileName: TFileName;
    FOptions: TAsSoundOptions;
    FSoundType: TAsSoundType;
  public
    procedure Assign(Source: TPersistent); override;
    constructor Create(AOwner: TComponent);
    destructor Destroy; override;
    function Play: Boolean;
  published
    property FileName: TFileName read FFileName write FFileName;
    property Options: TAsSoundOptions read FOptions write FOptions;
    property SoundType: TAsSoundType read FSoundType write FSoundType
      default asstFileName;
  end;
{$ENDIF}

  TAsSynAutoCorrectOption = (ascoCorrectOnMouseDown, ascoIgnoreCase,
    ascoMaintainCase, ascoSoundOnCorrect);
  TAsSynAutoCorrectOptions = set of TAsSynAutoCorrectOption;
               
  TAutoCorrectAction = (aaCorrect, aaAbort);
  TAutoCorrectEvent = procedure(Sender: TObject;
    const AOriginal, ACorrection: string; Line, Column: Integer;
    var Action: TAutoCorrectAction) of object;

  TCustomSynAutoCorrect = class(TComponent)
  private
    { Private declarations }

    { Published properties and events. }
    FEditor: TCustomSynEdit;
    FEnabled: Boolean;
    FItems: TStrings;
    FItemSepChar: Char;
    FOptions: TAsSynAutoCorrectOptions;
  {$IFDEF SOUND_SUPPORT}
    FSoundInfo: TAsSoundInfo;
  {$ENDIF}

    FOnAutoCorrect: TAutoCorrectEvent;

    { Private variables and methods. }
    FEditorEvents: PEditorEventsRec;
    FPrevLine: Integer;

    function CorrectItemStart(EditLine, SearchString: string; StartPos: LongInt;
      MatchCase, WholeWord: Boolean): LongInt;
    function FindAndCorrect(var EditLine: string; Original, Correction: string;
      var CurrentX: Integer): Boolean;
    function PreviousToken: string;

    { Accessor methods. }
    function GetItems: TStrings;
    procedure SetItems(const Value: TStrings);
  protected
    { Protected declarations }
    procedure EditorKeyDown(Sender: TObject; var Key: Word;
      Shift: TShiftState); virtual;
    procedure EditorMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer); virtual;
    procedure SetEditor(Value: TCustomSynEdit);
  public
    { Public declarations }
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;

    procedure Add(AOriginal, ACorrection: string);
    function AutoCorrectAll: Boolean;
    procedure Delete(AIndex: Integer);
    procedure Edit(AIndex: Integer; ANewOriginal, ANewCorrection: string);

  {$IFDEF INI_FILES}
    procedure LoadFromINI(AFileName, ASection: string);
    procedure SaveToINI(AFileName, ASection: string);
  {$ENDIF}

  {$IFDEF REGISTRY}
    procedure LoadFromRegistry(ARoot: DWORD; AKey: string);
    procedure SaveToRegistry(ARoot: DWORD; AKey: string);
  {$ENDIF}

    function LoadFromList(AFileName: string): Boolean;
    procedure SaveToList(AFileName: string);

    { Utility functions. }
    function HalfString(Str: string; GetFirstHalf: Boolean): string;  
  {$IFDEF OLD_SAVE_METHOD}
    function StrLeft(const s: string; Count: Integer): string;
    function StringsToStr(const List: TStrings; Sep: string): string;
    procedure StrToStrings(s: string; Sep: string; const List: TStrings);
  {$ENDIF}
  published
    { Published declarations }
    property Enabled: Boolean read FEnabled write FEnabled default True;
    property Editor: TCustomSynEdit read FEditor write SetEditor;
    property Items: TStrings read GetItems write SetItems;
    property ItemSepChar: Char read FItemSepChar write FItemSepChar default #9;
    property Options: TAsSynAutoCorrectOptions read FOptions write FOptions
      default [ascoIgnoreCase, ascoMaintainCase];
  {$IFDEF SOUND_SUPPORT}
    property SoundInfo: TAsSoundInfo read FSoundInfo write FSoundInfo;
  {$ENDIF}

    property OnAutoCorrect: TAutoCorrectEvent read FOnAutoCorrect
      write FOnAutoCorrect;
  end;

  TSynAutoCorrect = class(TCustomSynAutoCorrect)
  published                
    { Published declarations }
    property Enabled;
    property Editor;
    property Items;
    property ItemSepChar;
    property Options;
  {$IFDEF SOUND_SUPPORT}
    property SoundInfo;
  {$ENDIF}

    property OnAutoCorrect;
  end;

var
  AC_Delimiters: set of Byte = [33..255];
  AC_Numbers: set of Byte = [48..57];

implementation

{ TAsSoundInfo }

{$IFDEF SOUND_SUPPORT}
procedure TAsSoundInfo.Assign(Source: TPersistent);
begin
  if Source is TAsSoundInfo then
    with TAsSoundInfo(Source) do
    begin
      Self.FileName := FileName;
      Self.Options := Options;
      Self.SoundType := SoundType;
    end
    else
      inherited;
end;

constructor TAsSoundInfo.Create(AOwner: TComponent);
begin
  inherited Create;
  FFileName := '';
  FOptions := [assoAsync];
  FSoundType := asstBeep;
end;

destructor TAsSoundInfo.Destroy;
begin
  inherited;
end;

function TAsSoundInfo.Play: Boolean;
var
  o: Cardinal;

begin
  if FSoundType = asstBeep then
    Result := MessageBeep(0)
  else
  begin
    o := SND_SYNC;
    case FSoundType of
      asstAlias: o := o or SND_ALIAS;
      asstFileName: o := o or SND_FILENAME;
    end;
    if assoAsync in FOptions then o := o or SND_ASYNC;
    if assoNoDefault in FOptions then o := o or SND_NODEFAULT;
    if assoNoWait in FOptions then o := o or SND_NOWAIT;
    Result := PlaySound(PChar(FFileName), 0, o);
  end;
end;
{$ENDIF}

constructor TCustomSynAutoCorrect.Create(AOwner: TComponent);
var
  i: Integer;

begin
  inherited Create(AOwner);

  for i := 33 to 255 do if IsCharAlpha(Chr(i)) then Exclude(AC_Delimiters, i);
  FEditor := nil;
  FEnabled := True;
  FItems := TStringList.Create;
  FItemSepChar := #9;
  FOptions := [ascoIgnoreCase, ascoMaintainCase];
  FPrevLine := -1;
{$IFDEF SOUND_SUPPORT}
  FSoundInfo := TAsSoundInfo.Create(Self);
{$ENDIF}
end;

destructor TCustomSynAutoCorrect.Destroy;
begin
  FItems.Free;
{$IFDEF SOUND_SUPPORT}
  FSoundInfo.Free;
{$ENDIF}
  if Assigned(FEditorEvents) then Dispose(FEditorEvents);

  inherited;
end;

{ Utilities. }
function TCustomSynAutoCorrect.HalfString(Str: string;
  GetFirstHalf: Boolean): string;
var
  i: Integer;

begin
  i := LastDelimiter(FItemSepChar, Str);
  if i = 0 then i := Pred(MaxInt);

  if GetFirstHalf then
    Result := Copy(Str, 1, Pred(i))
  else
    Result := Copy(Str, Succ(i), MaxInt);
end;
             
{$IFDEF OLD_SAVE_METHOD}
function TCustomSynAutoCorrect.StrLeft(const s: string; Count: Integer): string;
begin
  Result := Copy(s, 1, Count);
end;

function TCustomSynAutoCorrect.StringsToStr(const List: TStrings;
  Sep: string): string;
var
  i, l: Integer;

begin
  Result := '';

  { Don't separate the two add operations! }
  for i := 0 to Pred(List.Count) do
  begin
    Result := Result + List[i];
    Result := Result + Sep;
  end;

  if List.Count > 0 then
  begin
    l := Length(Sep);
    System.Delete(Result, Succ(Length(Result) - l), l);
  end;
end;

procedure TCustomSynAutoCorrect.StrToStrings(s, Sep: string;
  const List: TStrings);
var
  i, l: Integer;
  Left: string;

begin
  List.Clear;
  l := Length(Sep);
  i := Pos(Sep, s);

  while i > 0 do
  begin
    Left := StrLeft(s, Pred(i));
    List.Add(Left);
    System.Delete(s, 1, Pred(i + l));
    I := Pos(Sep, s);
  end;
  if s <> '' then List.Add(s);
end;
{$ENDIF}

{$IFDEF INI_FILES}
procedure TCustomSynAutoCorrect.LoadFromIni(AFileName, ASection: string);
var
{$IFNDEF OLD_SAVE_METHOD}
  i: Integer;
  Original, Correction: string;
{$ENDIF}
  Reg: TIniFile;

begin
  Reg := TIniFile.Create(AFileName);
  try
  {$IFDEF OLD_SAVE_METHOD}
    StrToStrings(Reg.ReadString(ASection, 'AutoCorrect Items', ''),
      FStrSepChar, FItems);
  {$ELSE}
    FItems.Clear;
    with Reg do
      for i := 0 to Pred(ReadInteger(ASection, 'Count', 0)) do
      begin
        Original := ReadString(ASection, 'Original' + IntToStr(i), '');
        Correction := ReadString(ASection, 'Correction' + IntToStr(i), '');
        if not ((Original = '') and (Correction = '')) then
          FItems.Add(Original + FItemSepChar + Correction);
      end;
  {$ENDIF}
  finally
    Reg.Free;
  end;
end;

procedure TCustomSynAutoCorrect.SaveToIni(AFileName, ASection: string);
var         
{$IFNDEF OLD_SAVE_METHOD}
  i: Integer;
{$ENDIF}
  Reg: TIniFile;

begin
  Reg := TIniFile.Create(AFileName);
  try
  {$IFDEF OLD_SAVE_METHOD}
    Reg.WriteString(ASection, 'AutoCorrect Items',
      StringsToStr(FItems, FStrSepChar));
  {$ELSE}
    with Reg do
    begin
      WriteInteger(ASection, 'Count', FItems.Count);
      for i := 0 to Pred(FItems.Count) do
      begin
        WriteString(ASection, 'Original' + IntToStr(i),
          HalfString(FItems[i], True));
        WriteString(ASection, 'Correction' + IntToStr(i),
          HalfString(FItems[i], False));
      end;
    end;
  {$ENDIF}
  finally
    Reg.Free;
  end;
end;
{$ENDIF}

function TCustomSynAutoCorrect.LoadFromList(AFileName: string): Boolean;
begin
  Result := False;
  if FileExists(AFileName) then
  begin
    FItems.LoadFromFile(AFileName);
    Result := True;
  end;
end;

procedure TCustomSynAutoCorrect.SaveToList(AFileName: string);
begin
  FItems.SaveToFile(AFileName);
end;

{$IFDEF REGISTRY}
procedure TCustomSynAutoCorrect.LoadFromRegistry(ARoot: DWORD; AKey: string);
var
{$IFNDEF OLD_SAVE_METHOD}
  i: Integer;
  Original, Correction: string;
{$ENDIF}
  Reg: TRegIniFile;

begin
  Reg := TRegIniFile.Create('');
  try
    with Reg do
    begin
      RootKey := ARoot;
      OpenKey(AKey, True);
    {$IFDEF OLD_SAVE_METHOD}
      StrToStrings(ReadString('', 'AutoCorrect Items', ''), FStrSepChar,
        FItems);
    {$ELSE}
      FItems.Clear;
      for i := 0 to Pred(ReadInteger('', 'Count', 0)) do
      begin
        Original := ReadString('', 'Original' + IntToStr(i), '');
        Correction := ReadString('', 'Correction' + IntToStr(i), '');
        if not ((Original = '') and (Correction = '')) then
          FItems.Add(Original + FItemSepChar + Correction);
      end;
    {$ENDIF}
      CloseKey;
    end;
  finally
    Reg.Free;
  end;
end;

procedure TCustomSynAutoCorrect.SaveToRegistry(ARoot: DWORD; AKey: string);
var
{$IFNDEF OLD_SAVE_METHOD}
  i: Integer;
{$ENDIF}
  Reg: TRegIniFile;

begin
  Reg := TRegIniFile.Create('');
  try
    with Reg do
    begin
      RootKey := ARoot;
      OpenKey(AKey, True);
  {$IFDEF OLD_SAVE_METHOD}
      WriteString('', 'AutoCorrect Items', StringsToStr(FItems, FStrSepChar));
  {$ELSE}
      WriteInteger('', 'Count', FItems.Count);
      for i := 0 to Pred(FItems.Count) do
      begin
        WriteString('', 'Original' + IntToStr(i), HalfString(FItems[i], True));
        WriteString('', 'Correction' + IntToStr(i),
          HalfString(FItems[i], False));
      end;
  {$ENDIF}
      CloseKey;
    end;
  finally
    Reg.Free;
  end;
end;
{$ENDIF}

procedure TCustomSynAutoCorrect.Add(AOriginal, ACorrection: string);
begin
  FItems.Add(AOriginal + FItemSepChar + ACorrection);
end;

function TCustomSynAutoCorrect.AutoCorrectAll: Boolean;
var
  i, cx: Integer;
  s, Original, Correction, CurrText: string;

begin
  Result := False;
  if Assigned(Editor) then
  begin
    s := Editor.Lines.Text;
    cx := -1;

    for i := 0 to Pred(FItems.Count) do
    begin
      CurrText := FItems[i];
      Original := HalfString(CurrText, True);
      Correction := HalfString(CurrText, False);
      FindAndCorrect(s, Original, Correction, cx);
    end;
    Editor.Lines.Text := s;
  end;
end;

function TCustomSynAutoCorrect.CorrectItemStart(EditLine, SearchString: string;
  StartPos: LongInt; MatchCase, WholeWord: Boolean): LongInt;
var
  SearchCount, I: Integer;
  C: Char;
  CharMap: array [Char] of Char;
  CurBuf, Buf: PChar;
  BufLen: Integer;

const
  WordDelimiters: set of Char = [#0..#32];

  function FindNextWordStart(var BufPtr: PChar): Boolean;
  begin
    while (SearchCount > 0) and not (BufPtr^ in WordDelimiters) do
    begin
      Inc(BufPtr, 1);
      Dec(SearchCount);
    end;

    while (SearchCount > 0) and (BufPtr^ in WordDelimiters) do
    begin
      Inc(BufPtr, 1);
      Dec(SearchCount);
    end;

    Result := SearchCount >= 0;  
  end;

  function ScanText(var BufPtr: PChar): Boolean;
  var
     FirstWord: Boolean;
  begin
    Result := False;

    FirstWord := True;

    if WholeWord then
    begin
       while (SearchCount > 0) and (BufPtr^ in WordDelimiters) do
       begin
         Inc(BufPtr, 1);
         Dec(SearchCount);
       end;
    end;

    while SearchCount >= 0 do
    begin
      if WholeWord and (FirstWord = False) then
        if not FindNextWordStart(BufPtr) then Break;
      I := 0;
      while (CharMap[BufPtr[I]] = SearchString[I+1]) do
      begin
        Inc(I);
        if I >= Length(SearchString) then
        begin
          if (not WholeWord) or (SearchCount = 0) or
            (BufPtr[I] in WordDelimiters) then
          begin
            Result := True;
            Exit;
          end;
          Break;
        end;
      end;
      FirstWord := False;
      Inc(BufPtr);
      Dec(SearchCount);
    end;
  end;

begin
  Result := -1;
  BufLen := Length(EditLine);
  Buf := PChar(EditLine);

  if BufLen > 0 then
  begin
    SearchCount := succ(BufLen - StartPos - Length(SearchString));

    if (SearchCount >= 0) and (SearchCount <= BufLen) and
      (StartPos + SearchCount <= BufLen) then
    begin
      CurBuf := @Buf[StartPos];
      for C := Low(CharMap) to High(CharMap) do
      {$IFDEF SYN_CLX}
      begin
        if not MatchCase then
          CharMap[C] := UpCase(C)
        else
          CharMap[C] := C;
      end;
      {$ELSE}
        CharMap[C] := C;
      {$ENDIF}
      if not MatchCase then
      begin
      {$IFDEF SYN_CLX}
        SearchString := UpperCase(SearchString);
      {$ELSE}
        CharUpperBuff(PChar(@CharMap), SizeOf(CharMap));
        CharUpperBuff(@SearchString[1], Length(SearchString));
      {$ENDIF}
      end;
      if not ScanText(CurBuf) then
        CurBuf := nil
      else
      begin
        if CurBuf <> nil then
          Result := CurBuf - Buf;
      end;
    end;
  end;

  CurBuf := nil;
end;

procedure TCustomSynAutoCorrect.Delete(AIndex: Integer);
begin
  FItems.Delete(AIndex);
end;

procedure TCustomSynAutoCorrect.Edit(AIndex: Integer;
  ANewOriginal, ANewCorrection: string);
begin
  if AIndex > -1 then
    FItems[AIndex] := ANewOriginal + FItemSepChar + ANewCorrection;
end;

procedure TCustomSynAutoCorrect.EditorKeyDown(Sender: TObject;
  var Key: Word; Shift: TShiftState);
var             
  b: Boolean;
  i, cx: Integer;
  s, Original, Correction, CurrText: string;

begin
  { Call original KeyDown event. }
  if Assigned(TEditorEventsRec(FEditorEvents^).kd) then
    TEditorEventsRec(FEditorEvents^).kd(Sender, Key, Shift);

  if Assigned(Editor) and Enabled then
  begin
    FPrevLine := Editor.CaretY;
    if Key in [VK_Return, VK_Tab, VK_Space, 190] then
    begin
      b := False;
      s := PreviousToken;
      if s <> '' then
      begin
        cx := Editor.CaretX;
        for i := 0 to Pred(FItems.Count) do
        begin
          CurrText := FItems[i];
          Original := HalfString(CurrText, True);
          Correction := HalfString(CurrText, False);
          b := b or FindAndCorrect(s, Original, Correction, cx);
        end;
           
      {$IFDEF SOUND_SUPPORT}
        if b and (ascoSoundOnCorrect in FOptions) then FSoundInfo.Play;
      {$ENDIF}
      end;
    end;
  end;
end;

procedure TCustomSynAutoCorrect.EditorMouseDown(Sender: TObject;
  Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
var              
  Action: TAutoCorrectAction;
  b: Boolean;
  i, cx: Integer;
  s, Original, Correction, CurrText: string;

begin
  { Call original MouseDown method. }
  if Assigned(TEditorEventsRec(FEditorEvents^).md) then
    TEditorEventsRec(FEditorEvents^).md(Sender, Button, Shift, X, Y);

  if ascoCorrectOnMouseDown in FOptions then
  begin
    if Assigned(Editor) and Enabled and (FPrevLine <> -1) then
    begin
      b := False;
      s := Editor.Lines[Pred(FPrevLine)];
      cx := -1;

      for i := 0 to Pred(FItems.Count) do
      begin
        CurrText := FItems[i];
        Original := HalfString(CurrText, True);
        Correction := HalfString(CurrText, False);
        b := b or FindAndCorrect(s, Original, Correction, cx);
      end;

      if b then
      begin
        if Assigned(FOnAutoCorrect) then
        begin
          Action := aaCorrect;
          FOnAutoCorrect(Self, Editor.Lines[Pred(FPrevLine)], s, Editor.CaretY,
            0, Action);
          if Action = aaAbort then Exit;
        end;
        Editor.Lines[Pred(FPrevLine)] := s;

      {$IFDEF SOUND_SUPPORT}
        if (ascoSoundOnCorrect in FOptions) then FSoundInfo.Play;
      {$ENDIF}
      end;
    end;
  end;
end;

function TCustomSynAutoCorrect.FindAndCorrect(var EditLine: string;
  Original, Correction: string; var CurrentX: Integer): Boolean;
var
  StartPos: LongInt;
  EndPos: Integer;
  FoundText, ReplaceDefText: string;
  p: TPoint;
  Action: TAutoCorrectAction;

  function FirstCapCase(S: string): string;
  begin
    if S <> '' then
    begin
      s := LowerCase(s);
      s[1] := UpCase(s[1]);
    end;

    Result := s;
  end;

begin
  Result := False;
  ReplaceDefText := Correction;
  StartPos := 0;
  EndPos := Length(Original);

  if (Editor <> nil) and not (Editor.ReadOnly) then
  begin
    StartPos := CorrectItemStart(EditLine, Original, StartPos,
      not (ascoIgnoreCase in FOptions), True);

    while StartPos > -1 do
    begin
      if (ascoMaintainCase in FOptions) then
      begin
        Correction := ReplaceDefText;
        FoundText := Copy(EditLine,StartPos+1,EndPos);

        if FoundText = AnsiUpperCase(FoundText) then
          Correction := AnsiUpperCase(Correction)
        else
        begin
          if FoundText = AnsiLowerCase(FoundText) then
            Correction := AnsiLowerCase(Correction)
          else
          begin
            if FoundText = FirstCapCase(FoundText) then
              Correction := FirstCapCase(Correction);
          end;
        end;
      end;

      if CurrentX > - 1 then
      begin
        p := Editor.CaretXY;
        if Assigned(FOnAutoCorrect) then
        begin
          Action := aaCorrect;
          FOnAutoCorrect(Self, Original, Correction, P.y, P.x, Action);

          if Action = aaAbort then Break;
        end;

        Editor.BeginUpdate;

        try
          if p.x = 0 then
            Editor.BlockBegin := Point(p.x - 1 - EndPos, p.y)
          else
            Editor.BlockBegin := Point(p.x - EndPos, p.y);

          Editor.BlockEnd := p;
          p := Editor.BlockBegin;
          Editor.SelText := Correction;
          Result := True;
        finally
          Editor.EndUpdate;
        end;

        Break;
      end
      else
      begin
        Result := True;
        EditLine := Copy(EditLine, 1, StartPos) + Correction +
          Copy(EditLine, StartPos + EndPos + 1, MaxInt);
        Inc(StartPos, EndPos);
        StartPos := CorrectItemStart(EditLine, Original, StartPos,
          not (ascoIgnoreCase in FOptions), True);
      end;
    end;
  end;
end;
                      
function TCustomSynAutoCorrect.GetItems: TStrings;
begin
  Result := FItems;
end;

function TCustomSynAutoCorrect.PreviousToken: string;
var
  i, cx: Integer;

begin
  Result := Editor.LineText;
  cx := Editor.CaretX;
  i := Pred(cx);

  if i <= Length(Result) then
  begin
    while (i > 0) and (Result[i] > ' ') do Dec(i);
    Inc(i);
    Result := Copy(Result, i, cx - i);
  end
  else
    Result := '';
end;

procedure TCustomSynAutoCorrect.SetEditor(Value: TCustomSynEdit);
begin
  if FEditor <> Value then
  begin
    FEditor := Value;

    if Assigned(FEditor) then
    begin
      New(FEditorEvents);
      with FEditorEvents^ do
      begin
        kd := Editor.OnKeyDown;
        md := TAsCustomSynEdit(Editor).OnMouseDown;
      end;
      Editor.FreeNotification(Self);

      if not (csDesigning in ComponentState) then
      begin
        Editor.OnKeyDown := EditorKeyDown;
        TAsCustomSynEdit(Editor).OnMouseDown := EditorMouseDown;
      end;
    end;
  end;
end;

procedure TCustomSynAutoCorrect.SetItems(const Value: TStrings);
begin
  FItems.Assign(Value);
end;

end.
